#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

from fnmatch import fnmatch
import os
import yaml
import re

from shared_helpers.io import err, warn, info


class Config(object):
    def __init__(self, path=None, **config):
        # Injecting the path makes testing a million times easier.
        self._CONFIG_PATH = path if path else os.path.abspath(os.path.expanduser('~/.tvm.yaml'))

        if config:  # A configuration injected (for testing), do not bother loading.
            self._config = config
        else:
            self._config = self._set_safe_defaults()
            self._read_config()

        self._validate_config()

    def _set_safe_defaults(self):
        return {
            'record_storage_path': '~/sample',
            'physical_storage_paths': ['~/Downloads/tvm'],
            'ignore_paths': ['Thumbs', '.DS_Store', '^\._.*'],
            'rss_feeds': ['http://showrss.info/feeds/all.rss'],
            'api_keys': {'the_movie_db': None},
            'downloads': {
                'download_path': '~/Downloads/shows/',
                'cache_path': '~/.tvm-cache/',
                'max_video_quality': 720
            }
        }

    def _read_config(self):
        try:
            with open(self._CONFIG_PATH, 'rU') as stream:
                configuration = yaml.load(stream)

            for key, config in configuration.iteritems():
                self._config[key] = config

        except IOError:
            info('No .tvm.yaml file was found at $HOME. Using safe defaults.')

    def _validate_config(self):
        EXPECTED_CONFIG_KEYS = [
            'record_storage_path',
            'physical_storage_paths',
            'ignore_paths',
            'rss_feeds',
            'api_keys',
            'downloads'
        ]

        # General expected top-level keys.
        for key in EXPECTED_CONFIG_KEYS:
            if not self._config[key]:
                warn('No value was assigned to the key, "%s".' % key)

        # Specific API keys users have to specify.
        EXPECTED_API_KEYS = ['the_movie_db']
        for key in EXPECTED_API_KEYS:
            if key not in self._config['api_keys']:
                warn('%s API key is missing.' % key)
                continue

            if not self._config['api_keys'][key]:
                warn('No value was assigned to the api key, "%s".' % key)

        # Download (check command) related configuration.
        EXPECTED_DOWNLOAD_KEYS = ['download_path', 'cache_path', 'max_video_quality']
        VALID_VIDEO_QUALITY_P = ['240', '480', '720', '1080']

        for key in EXPECTED_DOWNLOAD_KEYS:
            if key not in self._config['downloads']:
                warn('%s not specified in downloads.' % key)
                continue

            # Not an expected video quality value.
            if key == 'max_video_quality' and str(self._config['downloads'][key]) not in VALID_VIDEO_QUALITY_P:
                warn('max_video_quality must be in %s.' % VALID_VIDEO_QUALITY_P)

    @property
    def record_storage_path(self):
        storage_path = self._config['record_storage_path'] if self._config['record_storage_path'] else '~/sample'
        return os.path.abspath(os.path.expanduser(storage_path))

    @property
    def physical_storage_paths(self):
        storage_path = self._config['physical_storage_paths']
        return map(lambda p: os.path.abspath(os.path.expanduser(p)), storage_path)

    @property
    def last_notification_path(self):
        return os.path.join(self.downloads['cache_path'], '.last_notification.json')

    @property
    def added_path(self):
        return os.path.join(self.record_storage_path, 'shows_added.csv')

    @property
    def viewed_path(self):
        return os.path.join(self.record_storage_path, 'shows_viewed.csv')

    @property
    def ignore_pattern(self):
        ignore_paths = self._config['ignore_paths'] if self._config['ignore_paths'] else []
        return re.compile('|'.join(ignore_paths))

    @property
    def api_keys(self):
        return self._config['api_keys']

    @property
    def downloads(self):
        # TODO: I should not need to re-expand these paths every time the property is called.
        for key, path in self._config['downloads'].iteritems():
            if fnmatch(key, '*_path'):
                path = self._config['downloads'][key]
                path = os.path.abspath(os.path.expanduser(path))
                self._config['downloads'][key] = path

        return self._config['downloads']
