#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import string

from tvm.shared_helpers.tvstatistics import TVStatistics
from tvm.shared_helpers.datehelpers import pretty_date
from tvm.shared_helpers.io import info, info2

from tvm.helpers import basehelper


class LsHelper(basehelper.BaseHelper):
    def __init__(self, config, collection):
        super(self.__class__, self).__init__(config, collection)

        self._statistics = TVStatistics(self.collection)

    def _calculate_max_column_width(self, table):
        # TODO: len should not treat ansi characters as part of the string. It screws up the positioning.
        column_widths = [0] * len(table[0])
        for i, column in enumerate(zip(*table)):
            for row in column:
                cell = str(row[1] if isinstance(row, (list, tuple)) else row)
                if len(cell) > column_widths[i]:
                    column_widths[i] = len(cell)
        return column_widths

    def _add_column_padding(self, table, padding=5):
        def _add_padding(column, i):
            return str(column) + ' ' * (column_widths[i] + padding - len(column))

        column_widths, new_table = self._calculate_max_column_width(table), []
        for row in table:
            new_row = []
            for index, column in enumerate(row):
                # Adjust center on all columns except first.
                adjust = string.center if index != 0 else string.ljust
                new_row.append(adjust(str(column), column_widths[index] + padding))

            new_table.append(new_row)
        return new_table

    def generate_table(self):
        table = [['title', 'last_updated', 'last_viewed', 'seen_latest?']]
        for title in sorted(self.collection.titles):
            statistics = self._statistics._gather_date_statistics(title)
            table.append([
                title,
                pretty_date(statistics['last_updated'][0]),
                pretty_date(statistics['last_viewed'][0]),
                "\033[0;32m✓\033[0m" if self.collection.viewed_up_to_date(title) else "\033[1;31m✖\033[0m"
            ])

        return table

    def generate_tree(self, title):
        table = []

        recorded, viewed = self.collection.find_all_episodes(title)
        grouped = self.collection.group_episodes_by_season(recorded)
        pending, incomplete, completed = self.collection.season_status(title=title)

        for season in sorted(map(int, grouped.keys())):
            episodes = grouped[str(season)]

            if str(season) in completed:
                table.append(['s%s' % season, '\033[0;32m✓\033[0m'])
                continue

            if str(season) in pending:
                table.append(['s%s' % season, '\033[1;31m✖\033[0m'])
                continue

            table.append(['s%s' % season, ''])
            for index, entry in enumerate(episodes):
                season, episode, title = entry
                episode_out = '--|- %s' if index == 0 else '  |- %s'
                table.append([
                    episode_out % episode,
                    "\033[0;32m✓\033[0m" if entry in viewed else "\033[1;31m✖\033[0m"
                ])

        return table

    def print_table(self, table):
        """ The columns of the first row in `table` are always the headings. """
        table = self._add_column_padding(table)

        info(''.join(table[0]))
        for i, row in enumerate(table[1:]):
            print('   ' + ''.join(map(str, row)))

    def print_tree(self, tree, title):
        """ The first column of the `tree` is a string representation of a tree
            but it's be honest, it's just a table.
        """
        tree = self._add_column_padding(tree)
        for i, row in enumerate(tree):
            info2(''.join(map(str, row)))

        recorded, viewed = self.collection.find_all_episodes(title)
        info('%d%% (%d/%d) viewed.' % ((round(float(len(viewed)) / len(recorded), 2) * 100),
                                       len(viewed), len(recorded)))
